from django.db.migrations.operations.special import RunSQL


class CreateTreeClosure(RunSQL):
    reversible = True

    def __init__(self, model_name):
        self.model_name = model_name.lower()

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        model = from_state.apps.get_model(app_label, self.model_name)
        view_name = model._meta.db_table + '_closure'

        schema_editor.execute('''
CREATE OR REPLACE RECURSIVE VIEW %(view)s(path, ancestor_id, descendant_id, depth) AS

SELECT ARRAY[node_id], node_id, node_id, 0
FROM %(table)s

UNION ALL

SELECT parent_id || path, parent_id, descendant_id, depth + 1
FROM %(table)s
INNER JOIN %(view)s ON (ancestor_id = node_id)
WHERE parent_id IS NOT NULL;
''' % {
    'table': schema_editor.connection.ops.quote_name(model._meta.db_table),
    'view': schema_editor.connection.ops.quote_name(view_name),
})


    def database_backwards(self, app_label, schema_editor, from_state, to_state):
        model = from_state.apps.get_model(app_label, self.model_name)
        view_name = model._meta.db_table + '_closure'
        schema_editor.execute('''DROP VIEW IF EXISTS %s;''' % (
            schema_editor.connection.ops.quote_name(view_name)
        ))



from django.db import models

from .fields import ClosureManyToManyField


class Node(models.Model):
    node_id = models.AutoField(primary_key=True)
    parent = models.ForeignKey('self', related_name='children', null=True, blank=True, on_delete=models.CASCADE)

    descendants = ClosureManyToManyField('self', symmetrical=False, related_name='ancestors')

    class Meta:
        abstract = True



from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.db.models.fields.related import resolve_relation
from django.db.models.utils import make_model_tuple
from django.utils.translation import gettext_lazy as _


class ClosureManyToManyField(models.ManyToManyField):
    '''
    Pre-configured M2M that defines a 'through' model automatically for the closure table.
    '''

    def contribute_to_class(self, cls, name, **kwargs):
        if not cls._meta.abstract:
            # Define through table
            meta = type('Meta', (), {
                'db_table': '%s_closure' % cls._meta.db_table,
                'app_label': cls._meta.app_label,
                'db_tablespace': cls._meta.db_tablespace,
                'unique_together': ('ancestor', 'descendant'),
                'verbose_name': _('ancestor-descendant relationship'),
                'verbose_name_plural': _('ancestor-descendant relationships'),
                'apps': cls._meta.apps,
                'managed': False,
            })
            # Construct and set the new class.
            name_ = '%s_Closure' % cls._meta.model_name
            self.remote_field.through = type(name_, (models.Model,), {
                'Meta': meta,
                '__module__': cls.__module__,
                'path': ArrayField(base_field=models.IntegerField(), primary_key=True),
                'ancestor': models.ForeignKey(
                    cls,
                    related_name='%s+' % name_,
                    db_tablespace=self.db_tablespace,
                    db_constraint=self.remote_field.db_constraint,
                    on_delete=models.DO_NOTHING,
                ),
                'descendant': models.ForeignKey(
                    cls,
                    related_name='%s+' % name_,
                    db_tablespace=self.db_tablespace,
                    db_constraint=self.remote_field.db_constraint,
                    on_delete=models.DO_NOTHING,
                ),
                'depth': models.IntegerField(),
            })

        super().contribute_to_class(cls, name, **kwargs)
